%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(binary_SUITE).

%% Tests binaries and the BIFs:
%%	list_to_binary/1
%%	binary_to_list/1
%%	binary_to_list/3
%%	binary_to_term/1
%%	term_to_binary/1
%%	size(Binary)
%%	concat_binary/1
%%	split_binary/2
%%      hash(Binary, N)
%%

-include("test_server.hrl").

-export([all/1, init_per_testcase/2, fin_per_testcase/2,
	 copy_terms/1, conversions/1, deep_lists/1,
	 bad_list_to_binary/1, bad_binary_to_list/1,
	 t_split_binary/1, bad_split/1, t_concat_binary/1, terms/1, terms_float/1,
	 t_hash/1,
	 bad_size/1, 
	 bad_binary_to_term_2/1,
	 bad_binary_to_term/1, bad_terms/1, bad_ref_and_port/1, more_bad_terms/1,
	 ordering/1,gc_test/1]).

%% Internal exports.
-export([sleeper/0]).

all(suite) -> [copy_terms,conversions, deep_lists,
	       t_split_binary, bad_split, t_concat_binary,
	       bad_list_to_binary, bad_binary_to_list, terms, terms_float,
	       bad_binary_to_term_2,
	       bad_binary_to_term, bad_terms, t_hash, bad_size,
	       bad_ref_and_port, more_bad_terms, ordering, gc_test].

init_per_testcase(Func, Config) when atom(Func), list(Config) ->
    Dog=?t:timetrap(?t:minutes(2)),
    [{watchdog, Dog}|Config].

fin_per_testcase(_Func, Config) ->
    Dog=?config(watchdog, Config),
    ?t:timetrap_cancel(Dog).

-define(heap_binary_size, 64).

copy_terms(Config) when list(Config) ->
    Self = self(),
    ?line Pid = spawn_link(fun() -> copy_server(Self) end),
    F = fun(Term) ->
		Pid ! Term,
		receive
		    Term -> ok;
		    Other ->
			io:format("Sent: ~P\nGot back:~P", [Term,12,Other,12]),
			?t:fail(bad_term)
		end
	end,
    ?line test_terms(F),
    ok.

copy_server(Parent) ->
    receive
	Term ->
	    Parent ! Term,
	    copy_server(Parent)
    end.

%% Tests list_to_binary/1, binary_to_list/1 and size/1,
%% using flat lists.

conversions(suite) -> [];
conversions(Config) when list(Config) ->
    ?line test_bin([]),
    ?line test_bin([1]),
    ?line test_bin([1, 2]),
    ?line test_bin([1, 2, 3]),
    ?line test_bin(lists:seq(0, ?heap_binary_size)),
    ?line test_bin(lists:seq(0, ?heap_binary_size+1)),
    ?line test_bin(lists:seq(0, 255)),
    ?line test_bin(lists:duplicate(50000, $@)),

    %% Binary in list.
    List = [1, 2, 3, 4, 5],
    ?line B1 = make_sub_binary(list_to_binary(List)),
    ?line 5 = size(B1),
    ?line B2 = list_to_binary([42, B1, 19]),
    ?line 7 = size(B2),
    ?line 7 = size(make_sub_binary(B2)),
    ?line [42, 1, 2, 3, 4, 5, 19] = binary_to_list(B2),
    ?line [42, 1, 2, 3, 4, 5, 19] = binary_to_list(make_sub_binary(B2)),

    ok.

test_bin(List) ->
    ?line Size = length(List),
    ?line Bin = list_to_binary(List),
    ?line Size = size(Bin),
    ?line Size = size(make_sub_binary(Bin)),
    ?line List = binary_to_list(Bin),
    ?line List = binary_to_list(make_sub_binary(Bin)).

%% Tests list_to_binary/1, binary_to_list/1,3 and size/1,
%% using deep lists.

deep_lists(suite) -> [];
deep_lists(Config) when list(Config) ->
    ?line test_deep_list(["abc"]),
    ?line test_deep_list([[12, 13, [123, 15]]]),
    ?line test_deep_list([[12, 13, [lists:seq(0, 255), []]]]),
    ok.

test_deep_list(List) ->
    ?line Flat_List = lists:flatten(List),
    ?line Size = length(Flat_List),
    ?line Bin = list_to_binary(List),
    ?line Size = size(Bin),
    ?line Flat_List = binary_to_list(Bin),
    ?line t_binary_to_list_3(Flat_List, Bin, 1, Size).

t_binary_to_list_3(List, Bin, From, To) ->
    ?line going_up(List, Bin, From, To),
    ?line going_down(List, Bin, From, To),
    ?line going_center(List, Bin, From, To).

going_up(List, Bin, From, To) when From =< To ->
    ?line List = binary_to_list(Bin, From, To),
    ?line going_up(tl(List), Bin, From+1, To);
going_up(_List, _Bin, From, To) when From > To ->
    ok.
    
going_down(List, Bin, From, To) when To > 0->
    ?line compare(List, binary_to_list(Bin, From, To), To-From+1),
    ?line going_down(List, Bin, From, To-1);
going_down(_List, _Bin, _From, _To) ->
    ok.

going_center(List, Bin, From, To) when From >= To ->
    ?line compare(List, binary_to_list(Bin, From, To), To-From+1),
    ?line going_center(tl(List), Bin, From+1, To-1);
going_center(_List, _Bin, _From, _To) ->
    ok.

compare([X|Rest1], [X|Rest2], Left) when Left > 0 ->
    ?line compare(Rest1, Rest2, Left-1);
compare([_X|_], [_Y|_], _Left) ->
    ?line test_server:fail();
compare(_List, [], 0) ->
    ok.

bad_list_to_binary(suite) -> [];
bad_list_to_binary(Config) when list(Config) ->
    ?line test_bad_bin(atom),
    ?line test_bad_bin(42),
    ?line test_bad_bin([1|2]),
    ?line test_bad_bin([256]),
    ?line test_bad_bin([255, [256]]),
    ?line test_bad_bin([-1]),
    ?line test_bad_bin([atom_in_list]),

    %% Funs are implmented as a type of binary internally.
    ?line test_bad_bin(fun(X, Y) -> X*Y end),
    ?line test_bad_bin([1,fun(X) -> X + 1 end,2|fun() -> 0 end]),
    ?line test_bad_bin([fun(X) -> X + 1 end]),
    ok.

test_bad_bin(List) ->
    ?line {'EXIT', {badarg, _}} = (catch list_to_binary(List)).

bad_binary_to_list(doc) -> "Tries binary_to_list/1,3 with bad arguments.";
bad_binary_to_list(suite) -> [];
bad_binary_to_list(Config) when list(Config) ->
    ?line bad_bin_to_list(fun(X) -> X * 42 end),

    GoodBin = list_to_binary(lists:seq(1, 10)),
    ?line bad_bin_to_list(fun(X) -> X * 44 end, 1, 2),
    ?line bad_bin_to_list(GoodBin, 0, 1),
    ?line bad_bin_to_list(GoodBin, 2, 1),
    ?line bad_bin_to_list(GoodBin, 11, 11),
    ok.

bad_bin_to_list(BadBin) ->
    {'EXIT',{badarg,_}} = (catch binary_to_list(BadBin)).

bad_bin_to_list(Bin, First, Last) ->
    {'EXIT',{badarg,_}} = (catch binary_to_list(Bin, First, Last)).
    
    
%% Tries to split a binary at all possible positions.

t_split_binary(suite) -> [];
t_split_binary(Config) when list(Config) ->
    ?line L = lists:seq(0, ?heap_binary_size-5), %Heap binary.
    ?line B = list_to_binary(L),
    ?line split(L, B, size(B)),

    %% Sub binary of heap binary.
    ?line split(L, make_sub_binary(B), size(B)),
    {X,_Y} = split_binary(B, size(B) div 2),
    ?line split(binary_to_list(X), X, size(X)),
    
    %% Reference-counted binary.
    ?line L2 = lists:seq(0, ?heap_binary_size+1),
    ?line B2 = list_to_binary(L2),
    ?line split(L2, B2, size(B2)),

    %% Sub binary of reference-counted binary.
    ?line split(L2, make_sub_binary(B2), size(B2)),
    {X2,_Y2} = split_binary(B2, size(B2) div 2),
    ?line split(binary_to_list(X2), X2, size(X2)),

    ok.

split(L, B, Pos) when Pos > 0 ->
    ?line {B1, B2} = split_binary(B, Pos),
    ?line B1 = list_to_binary(lists:sublist(L, 1, Pos)),
    ?line B2 = list_to_binary(lists:nthtail(Pos, L)),
    ?line split(L, B, Pos-1);
split(_L, _B, 0) ->
    ok.

bad_split(doc) -> "Tries split_binary/2 with bad arguments.";
bad_split(suite) -> [];
bad_split(Config) when list(Config) ->
    GoodBin = list_to_binary([1,2,3]),
    ?line bad_split(GoodBin, -1),
    ?line bad_split(GoodBin, 4),
    ?line bad_split(GoodBin, a),

    %% Funs are a kind of binaries.
    ?line bad_split(fun(_X) -> 1 end, 1),
    ok.
    
bad_split(Bin, Pos) ->
    {'EXIT',{badarg,_}} = (catch split_binary(Bin, Pos)).

%% Tests concat_binary/2 and size/1.

t_concat_binary(suite) -> [];
t_concat_binary(Config) when list(Config) ->
    test_concat([]),

    test_concat([[]]),
    test_concat([[], []]),
    test_concat([[], [], []]),

    test_concat([[1], []]),
    test_concat([[], [2]]),
    test_concat([[], [3], []]),

    test_concat([[1, 2, 3], [4, 5, 6, 7]]),
    test_concat([[1, 2, 3], [4, 5, 6, 7], [9, 10]]),

    test_concat([lists:seq(0, 255), lists:duplicate(1024, $@),
		 lists:duplicate(2048, $a),
		 lists:duplicate(4000, $b)]),
    ok.

test_concat(Lists) ->
    test_concat(Lists, 0, [], []).

test_concat([List|Rest], Size, Combined, Binaries) ->
    ?line Bin = list_to_binary(List),
    ?line test_concat(Rest, Size+length(List), Combined++List, [Bin|Binaries]);
test_concat([], Size, Combined, Binaries0) ->
    ?line Binaries = lists:reverse(Binaries0),
    ?line Bin = concat_binary(Binaries),
    ?line Size = size(Bin),
    ?line Combined = binary_to_list(Bin).

t_hash(doc) -> "Test hash/2 with different type of binaries.";
t_hash(Config) when list(Config) ->
    ?line test_hash([]),
    ?line test_hash([253]),
    ?line test_hash(lists:seq(1, ?heap_binary_size)),
    ?line test_hash(lists:seq(1, ?heap_binary_size+1)),
    ?line test_hash([X rem 256 || X <- lists:seq(1, 312)]),
    ok.

test_hash(List) ->
    N = 65535,
    Bin = list_to_binary(List),
    Sbin = make_sub_binary(List),
    case {erlang:hash(Bin, N),erlang:hash(Sbin, N)} of
	{H,H} -> ok;
	{H1,H2} ->
	    io:format("Different hash values: ~p, ~p\n", [H1,H2]),
	    ?t:fail()
    end,
    case {erlang:phash(Bin, N),erlang:phash(Sbin, N)} of
	{PH,PH} -> ok;
	{PH1,PH2} ->
	    io:format("Different hash values: ~p, ~p\n", [PH1,PH2]),
	    ?t:fail()
    end,
    case {erlang:phash2(Bin, N),erlang:phash2(Sbin, N)} of
	{PH_2,PH_2} -> ok;
	{PH1_2,PH2_2} ->
	    io:format("Different hash values: ~p, ~p\n", [PH1_2,PH2_2]),
	    ?t:fail()
    end.


bad_size(doc) -> "Try bad arguments to size/1.";
bad_size(suite) -> [];
bad_size(Config) when list(Config) ->
    ?line {'EXIT',{badarg,_}} = (catch size(fun(X) -> X + 33 end)),
    ok.
    
%% Tests that the emulator doesn't crash when Corrupt references and ports are
%% received or decoded from the extern format
%% This test will cause the emulator to crash (core dump) for version 4.7.3 and
%% earlier
bad_ref_and_port(suite) ->[];
bad_ref_and_port(Config) when list(Config)->
    ?line Port = make_port(),
    ?line ok = test_ref_and_port(Port),
% Don't test refs in this way for the time being, since the format
% has changed.
%    ?line Ref = make_ref(),
%    ?line ok = test_ref_and_port(Ref),
    ok.

test_ref_and_port(Ref) ->
    ?line Bin = term_to_binary(Ref),
    ?line Ref = binary_to_term(Bin),
    List = binary_to_list(Bin),
    [Creat,N0,N1,N2,N3|Front] = lists:reverse(List),
    %% Creat is set to something else then 'this_creation' 
    %% to see if Creat is preserved during decode and encode
%    NewCreat = case Creat + 1 of
%		   _Nc when _Nc =< 3 ->
%		       _Nc;
%		   _ -> 3
%	       end,
    List1 = lists:reverse(Front) ++ [N3,N2,N1,N0,3],
    List2 = binary_to_list(term_to_binary(binary_to_term(list_to_binary(List1)))),
    io:format("Before: ~p~nAfter: ~p~n",[List1,List2]),
    ?line List1 = List2,
    %% now changing the ref to be above the maximum 16#400-1 (i.e 18 bits)
    NewList = lists:reverse(Front) ++ [0,4,0,0,Creat],
    %% the original ref
    NewBin = list_to_binary(NewList),
    ?line {'EXIT',{badarg,_}} = (catch binary_to_term(NewBin)),
    ok.

%% Tests binary_to_term/1 and term_to_binary/1.

terms(suite) -> [];
terms(Config) when list(Config) ->
    TestFun = fun(Term) ->
		      io:format("~p", [Term]),
		      Bin = term_to_binary(Term),
		      Term = binary_to_term(Bin),
		      BinC = erlang:term_to_binary(Term, [compressed]),
		      Term = binary_to_term(BinC)
	      end,
    ?line test_terms(TestFun),
    ok.

terms_float(suite) -> [];
terms_float(Config) when list(Config) ->
    ?line test_floats(fun(Term) ->
			      Bin = term_to_binary(Term),
			      Term = binary_to_term(Bin)
		      end).

bad_binary_to_term_2(doc) -> "OTP-4053.";
bad_binary_to_term_2(suite) -> [];
bad_binary_to_term_2(Config) when list(Config) ->
    ?line {ok, N} = test_server:start_node(plopp, slave, []),
    ?line R = rpc:call(N, erlang, binary_to_term, [<<131,111,255,255,255,0>>]),
    ?line case R of
	      {badrpc, {'EXIT', _}} ->
		  ok;
	      _Other ->
		  test_server:fail({rpcresult, R})
	  end,
    ?line test_server:stop_node(N),
    ok.

bad_binary_to_term(doc) -> "Try bad input to binary_to_term/1.";
bad_binary_to_term(suite) -> [];
bad_binary_to_term(Config) when list(Config) ->
    ?line bad_bin_to_term(an_atom),
    ?line bad_bin_to_term({an,tuple}),
    ?line bad_bin_to_term({a,list}),
    ?line bad_bin_to_term(fun() -> self() end),
    ?line bad_bin_to_term(fun(X) -> 42*X end),
    ?line bad_bin_to_term(fun(X, Y) -> {X,Y} end),
    ?line bad_bin_to_term(fun(X, Y, Z) -> {X,Y,Z} end),
    ok.

bad_bin_to_term(BadBin) ->
    {'EXIT',{badarg,_}} = (catch binary_to_term(BadBin)).

%% Tests bad input to binary_to_term/1.

bad_terms(suite) -> [];
bad_terms(Config) when list(Config) ->
    ?line test_terms(fun corrupter/1).
			     
corrupter(Term) ->
%%    ?line test_server:purify_format("About to corrupt: ~P", [Term,12]),
    ?line io:format("About to corrupt: ~P", [Term,12]),
    ?line Bin = term_to_binary(Term),
    ?line corrupter(Bin, size(Bin)-1),
    ?line CompressedBin = term_to_binary(Term, [compressed]),
    ?line corrupter(CompressedBin, size(CompressedBin)-1).

corrupter(Bin, Pos) when Pos >= 0 ->
    ?line {ShorterBin, _} = split_binary(Bin, Pos),
    ?line catch binary_to_term(ShorterBin), %% emulator shouldn't crash
    ?line MovedBin = list_to_binary([ShorterBin]),
    ?line catch binary_to_term(MovedBin), %% emulator shouldn't crash
    ?line corrupter(MovedBin, Pos-1);
corrupter(_Bin, _) ->
    ok.

more_bad_terms(suite) -> [];
more_bad_terms(Config) when list(Config) ->
    ?line Data = ?config(data_dir, Config),
    ?line BadFile = filename:join(Data, "bad_binary"),
    ?line ok = io:format("File: ~s\n", [BadFile]),
    ?line case file:read_file(BadFile) of
	      {ok,Bin} ->
		  ?line {'EXIT',{badarg,_}} = (catch binary_to_term(Bin)),
		  ok;
	      Other ->
		  ?line ?t:fail(Other)
	  end.

ordering(doc) -> "Tests ordering of binaries.";
ordering(suite) -> [];
ordering(Config) when list(Config) ->
    B1 = list_to_binary([7,8,9]),
    B2 = make_sub_binary([1,2,3,4]),
    B3 = list_to_binary([1,2,3,5]),

    %% From R8 binaries are compared as strings.

    ?line false = B1 == B2,
    ?line false = B1 =:= B2,
    ?line true = B1 /= B2,
    ?line true = B1 =/= B2,

    ?line true = B1 > B2,
    ?line true = B2 < B3,
    ?line true = B2 =< B1,
    ?line true = B2 =< B3,

    %% Binaries are greater than all other terms.

    ?line true = B1 > 0,
    ?line true = B1 > 39827491247298471289473333333333333333333333333333,
    ?line true = B1 > -3489274937438742190467869234328742398347,
    ?line true = B1 > 3.14,
    ?line true = B1 > [],
    ?line true = B1 > [a],
    ?line true = B1 > {a},
    ?line true = B1 > self(),
    ?line true = B1 > make_ref(),
    ?line true = B1 > xxx,
    ?line true = B1 > fun() -> 1 end,

    ?line Path = ?config(priv_dir, Config),
    ?line AFile = filename:join(Path, "vanilla_file"),
    ?line Port = open_port(AFile, [out]),
    ?line true = B1 > Port,

    ?line true = B1 >= 0,
    ?line true = B1 >= 39827491247298471289473333333333333333333333333333,
    ?line true = B1 >= -3489274937438742190467869234328742398347,
    ?line true = B1 >= 3.14,
    ?line true = B1 >= [],
    ?line true = B1 >= [a],
    ?line true = B1 >= {a},
    ?line true = B1 >= self(),
    ?line true = B1 >= make_ref(),
    ?line true = B1 >= xxx,
    ?line true = B1 >= fun() -> 1 end,
    ?line true = B1 >= Port,

    ok.
    
test_terms(Test_Func) ->
    ?line Test_Func(atom),
    ?line Test_Func(''),
    ?line Test_Func('a'),
    ?line Test_Func('ab'),
    ?line Test_Func('abc'),
    ?line Test_Func('abcd'),
    ?line Test_Func('abcde'),
    ?line Test_Func('abcdef'),
    ?line Test_Func('abcdefg'),
    ?line Test_Func('abcdefgh'),

    ?line Test_Func({trace_ts,{even_bigger,{some_data,fun() -> ok end}},{1,2,3}}),
    ?line Test_Func({trace_ts,{even_bigger,{some_data,<<1,2,3,4,5,6,7,8,9,10>>}},
		     {1,2,3}}),

    ?line Test_Func(1),
    ?line Test_Func(42),
    ?line Test_Func(-23),
    ?line Test_Func(256),
    ?line Test_Func(25555),
    ?line Test_Func(-3333),

    ?line Test_Func(1.0),

    ?line Test_Func(183749783987483978498378478393874),
    ?line Test_Func(-37894183749783987483978498378478393874),
    Very_Big = very_big_num(),
    ?line Test_Func(Very_Big),
    ?line Test_Func(-Very_Big+1),

    ?line Test_Func([]),
    ?line Test_Func("abcdef"),
    ?line Test_Func([a, b, 1, 2]),
    ?line Test_Func([a|b]),

    ?line Test_Func({}),
    ?line Test_Func({1}),
    ?line Test_Func({a, b}),
    ?line Test_Func({a, b, c}),
    ?line Test_Func(list_to_tuple(lists:seq(0, 255))),
    ?line Test_Func(list_to_tuple(lists:seq(0, 256))),

    ?line Test_Func(make_ref()),
    ?line Test_Func([make_ref(), make_ref()]),

    ?line Test_Func(make_port()),

    ?line Test_Func(make_pid()),

    ?line Test_Func(list_to_binary(lists:seq(0, 14))),
    ?line Test_Func(list_to_binary(lists:seq(0, ?heap_binary_size))),
    ?line Test_Func(list_to_binary(lists:seq(0, ?heap_binary_size+1))),
    ?line Test_Func(list_to_binary(lists:seq(0, 255))),

    ?line Test_Func(make_sub_binary(lists:seq(42, 43))),
    ?line Test_Func(make_sub_binary([42,43,44])),
    ?line Test_Func(make_sub_binary([42,43,44,45])),
    ?line Test_Func(make_sub_binary([42,43,44,45,46])),
    ?line Test_Func(make_sub_binary([42,43,44,45,46,47])),
    ?line Test_Func(make_sub_binary([42,43,44,45,46,47,48])),
    ?line Test_Func(make_sub_binary(lists:seq(42, 49))),
    ?line Test_Func(make_sub_binary(lists:seq(0, 14))),
    ?line Test_Func(make_sub_binary(lists:seq(0, ?heap_binary_size))),
    ?line Test_Func(make_sub_binary(lists:seq(0, ?heap_binary_size+1))),
    ?line Test_Func(make_sub_binary(lists:seq(0, 255))),

    ?line Test_Func(F = fun(A) -> 42*A end),
    ?line Test_Func(lists:duplicate(32, F)),

    ok.

test_floats(Test_Func) ->
    ?line Test_Func(5.5),
    ?line Test_Func(-15.32),
    ?line Test_Func(1.2435e25),
    ?line Test_Func(1.2333e-20),
    ?line Test_Func(199.0e+15),
    ok.

very_big_num() ->
    very_big_num(33, 1).

very_big_num(Left, Result) when Left > 0 ->
    ?line very_big_num(Left-1, Result*256);
very_big_num(0, Result) ->
    ?line Result.

make_port() ->
    ?line open_port({spawn, efile}, [eof]).

make_pid() ->
    ?line spawn_link(?MODULE, sleeper, []).

sleeper() ->
    ?line receive after infinity -> ok end.


gc_test(doc) -> "Test that binaries are garbage collected properly.";
gc_test(suite) -> [];
gc_test(Config) when list(Config) ->
    %% Note: This test is only relevant for REFC binaries.
    %% Therefore, we take care that all binaries are REFC binaries.

    B = list_to_binary(lists:seq(0, ?heap_binary_size)),
    Self = self(),
    F1 = fun() ->
		 gc(),
		 {binary,[]} = process_info(self(), binary),
		 Self ! {self(),done}
	 end,
    F = fun() ->
		receive go -> ok end,
		{binary,[{_,65,1}]} = process_info(self(), binary),
		gc(),
		{B1,B2} = my_split_binary(B, 4),
		gc(),
		gc(),
		{binary,L1} = process_info(self(), binary),
		[Binfo1,Binfo2,Binfo3] = L1,
		{_,65,3} = Binfo1 = Binfo2 = Binfo3,
		65 = size(B),
		4 = size(B1),
		61 = size(B2),
		F1()
	end,
    gc(),
    gc(),
    65 = size(B),
    gc_test1(spawn_opt(erlang, apply, [F,[]], [link,{fullsweep_after,0}])).

gc_test1(Pid) ->
    gc(),
    Pid ! go,
    receive
	{Pid,done} -> ok
    after 10000 ->
	    ?line ?t:fail()
    end.

%% Like split binary, but returns REFC binaries. Only useful for gc_test/1.

my_split_binary(B, Pos) ->
    Self = self(),
    Ref = make_ref(),
    spawn(fun() -> Self ! {Ref,split_binary(B, Pos)} end),
    receive
	{Ref,Result} -> Result
    end.

gc() ->
    erlang:garbage_collect(),
    gc1().
gc1() -> ok.

make_sub_binary(Bin) when binary(Bin) ->
    {_,B} = split_binary(list_to_binary([0,1,3,Bin]), 3),
    B;
make_sub_binary(List) ->
    make_sub_binary(list_to_binary(List)).
