%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(fun_r7_SUITE).

-define(default_timeout, ?t:minutes(1)).

-export([all/1,init_per_testcase/2,fin_per_testcase/2,dist_r7/1]).

-include("test_server.hrl").

all(suite) -> [dist_r7].

init_per_testcase(Case, Config) ->
    ?line Dog = test_server:timetrap(?default_timeout),
    [{watchdog, Dog}|Config].

fin_per_testcase(Case, Config) ->
    Dog=?config(watchdog, Config),
    test_server:timetrap_cancel(Dog),
    ok.

dist_r7(Config) when list(Config) ->
    case os:type() of
	{unix,sunos} ->
	    do_dist_r7(Config);
	_ ->
	    {skipped,"Works on {unix,sunos} only"}
    end.

do_dist_r7(Config) when list(Config) ->
    ?line Pa = filename:dirname(code:which(?MODULE)),
    Name = fun_dist_r7,
    ?line {ok,Node} = ?t:start_node(Name, peer,
				    [{args,"-pa "++Pa},
				     {erl,[{release,"r7b_patched"}]}]),

    ?line Pid = spawn_link(Node,
			   fun() ->
				   receive
				       Fun when function(Fun) ->
					   Fun(Fun)
				   end
			   end),
    Self = self(),
    Fun = fun(F) ->
		  Self =:= erlang:fun_info(F, pid),
		  ?MODULE =:= erlang:fun_info(F, module),
		  Self ! {ok,F}
	  end,
    ?line Pid ! Fun,
    ?line receive
	      {ok,Fun} -> ok;
	      Other -> ?t:fail({bad_message,Other})
	  end,
    ok.
