#include "sl_alloc_test.h"
#include <sys/wait.h>

#ifdef USE_THREADS

#ifndef THREAD_SAFE_SL_ALLOC
#error "*** sl_alloc not thread safe ***"
#endif

#define NO_OF_ALLOC_SEQS 6
#define NO_OF_THREADS (2*NO_OF_ALLOC_SEQS)
#define NO_OF_BLOCKS 10
#define NO_OF_OPS_PER_BL 200
#define FORK_INTERVALL   30
#define SBC_THRESHOLD 8192

#define BLOCK_ID(TN, BN) ((TN) << 4 | (BN))

Uint alloc_seq_1[] = {
  17,
  SBC_THRESHOLD*2,
  SBC_THRESHOLD*20,
  SBC_THRESHOLD*2,
  17,
  0
};

Uint alloc_seq_2[] = {
  SBC_THRESHOLD*20,
  SBC_THRESHOLD*2,
  17,
  SBC_THRESHOLD*2,
  SBC_THRESHOLD*20,
  0
};

Uint alloc_seq_3[] = {
  1,
  SBC_THRESHOLD/10,
  SBC_THRESHOLD/9,
  SBC_THRESHOLD/8,
  SBC_THRESHOLD/7,
  SBC_THRESHOLD/6,
  SBC_THRESHOLD/5,
  SBC_THRESHOLD/4,
  SBC_THRESHOLD/3,
  SBC_THRESHOLD/2,
  SBC_THRESHOLD/1,
  SBC_THRESHOLD*1,
  SBC_THRESHOLD*2,
  SBC_THRESHOLD*3,
  SBC_THRESHOLD*4,
  SBC_THRESHOLD*5,
  SBC_THRESHOLD*6,
  SBC_THRESHOLD*7,
  SBC_THRESHOLD*8,
  SBC_THRESHOLD*9,
  SBC_THRESHOLD*10,
  SBC_THRESHOLD*9,
  SBC_THRESHOLD*8,
  SBC_THRESHOLD*7,
  SBC_THRESHOLD*6,
  SBC_THRESHOLD*5,
  SBC_THRESHOLD*4,
  SBC_THRESHOLD*3,
  SBC_THRESHOLD*2,
  SBC_THRESHOLD*1,
  SBC_THRESHOLD/2,
  SBC_THRESHOLD/3,
  SBC_THRESHOLD/4,
  SBC_THRESHOLD/5,
  SBC_THRESHOLD/6,
  SBC_THRESHOLD/7,
  SBC_THRESHOLD/8,
  SBC_THRESHOLD/9,
  SBC_THRESHOLD/10,
  1,
  0
};

Uint alloc_seq_4[] = {
  SBC_THRESHOLD*2,
  SBC_THRESHOLD*3,
  SBC_THRESHOLD*7,
  SBC_THRESHOLD*8,
  SBC_THRESHOLD*5,
  SBC_THRESHOLD*6,
  SBC_THRESHOLD*1,
  SBC_THRESHOLD*10,
  SBC_THRESHOLD*4,
  SBC_THRESHOLD*2,
  0
};

Uint alloc_seq_5[] = {
  SBC_THRESHOLD/2,
  SBC_THRESHOLD/3,
  SBC_THRESHOLD/7,
  SBC_THRESHOLD/8,
  SBC_THRESHOLD/5,
  SBC_THRESHOLD/6,
  SBC_THRESHOLD/1,
  SBC_THRESHOLD/10,
  SBC_THRESHOLD/4,
  SBC_THRESHOLD/2,
  SBC_THRESHOLD/3,
  SBC_THRESHOLD/7,
  SBC_THRESHOLD/8,
  SBC_THRESHOLD/5,
  SBC_THRESHOLD/6,
  SBC_THRESHOLD/1,
  SBC_THRESHOLD/10,
  SBC_THRESHOLD/4,
  SBC_THRESHOLD/2,
  0
};

Uint alloc_seq_6[] = {
    1, 50, 100, 50, 23, 46, 2345, 23, 54, 2, 0
};

Uint *alloc_seqs[NO_OF_ALLOC_SEQS] = {
  alloc_seq_1,
  alloc_seq_2,
  alloc_seq_3,
  alloc_seq_4,
  alloc_seq_5,
  alloc_seq_6
};

static int no_fork_test = 0;

void
do_fork(int thr_no)
{
  pid_t res;
  int status;
  pid_t pid;

  if(no_fork_test)
    return;

  printf("Thread %d about to fork()\n", thr_no);
  fflush(stdout);
  pid = fork();

  if(!pid) {
    /* Child ... */

    /* If erts_sl_alloc() isn't forksafe (which it is) this can cause
       deadlock in the child */
    void *p = erts_sl_alloc(SBC_THRESHOLD*2);
    if(!p)
      exit(1);
    erts_sl_free(p);
    exit(17);
  }

  /* Parent ... */
  printf("Thread %d fork()ed\n", thr_no);
  fflush(stdout);
  if(pid < 0)
    FAIL_TEST;

  do {
    res = waitpid(pid, &status, 0);
  } while (res < 0 && errno == EINTR);

  if(res != pid) {
    printf("errno = %d\n", errno);
    FAIL_TEST;
  }

  if(!(WIFEXITED(status) && WEXITSTATUS(status) == 17)) {
    printf("status = %d\n", status);
    FAIL_TEST;
  }

  printf("Thread %d's child dead\n", thr_no);
  fflush(stdout);
}

void *
thread_func(void *arg)
{
  int i, j;
  int tn = *((int *) arg);
  block bs[NO_OF_BLOCKS];

  printf("Thread no %d started!\n", tn);
  fflush(stdout);

  for(i = 0; i < NO_OF_BLOCKS; i++) {
    bs[i].p = NULL;
    bs[i].s = 0;
    bs[i].i = 0;
    bs[i].as = alloc_seqs[i % NO_OF_ALLOC_SEQS];
  }

  for(i = 0; i < NO_OF_OPS_PER_BL; i++) {

    if((i + tn) % (FORK_INTERVALL) == 0)
      do_fork(tn);

    for(j = 0; j < NO_OF_BLOCKS; j++)
      alloc_op(&bs[j], BLOCK_ID(tn, j), 0);

  }

  for(j = 0; j < NO_OF_BLOCKS; j++)
    alloc_op(&bs[j], BLOCK_ID(tn, j), 1);

  printf("Thread no %d ended!\n", tn);
  fflush(stdout);
  return NULL;

}

int thread_arg[NO_OF_THREADS];
erts_thread_t threads[NO_OF_THREADS];

START_TEST
{
  int i;
  ErtsSlAllocInit esai;

  memset((void *) &esai, 255, sizeof(ErtsSlAllocInit));
  esai.esla = 1;
  esai.eosla = !test_new_sl_alloc;
  esai.sbct = SBC_THRESHOLD;
  esai.mmc = 5000;

  erts_sl_alloc_init(&esai);

  for(i = 0; i < NO_OF_THREADS; i++) {
    int res;
    thread_arg[i] = i + 1;
    res = erts_thread_create(&threads[i],
			     thread_func,
			     (void *) &thread_arg[i],
			     0);
    if(0 != res) {
      printf("Failed to create thread %d (error_code = %d)\n", i + 1, res);
      FAIL_TEST;
    }
  }

  for(i = 0; i < NO_OF_THREADS; i++) {
    int res;
    void *exit_value;
    res = erts_thread_join(threads[i], &exit_value);
    if(res != 0) {
      printf("Failed to join thread %d (error_code = %d)\n", i + 1, res);
      FAIL_TEST;
    }
  }

  TEST_SUCCEEDED;
}
END_TEST

#else

START_TEST
{
  SKIP_TEST("Threads not used");
}
END_TEST

#endif
